/**
 * Maa ewolucja - Gra ewolucyjno spoeczna
 *
 *	Gwna klasa odpowiadajca za kontrole przebiegiem symulacji oraz ca logik aplikacji.
 *
 * @author Grzegorz Przytua, Adam Makowski
 */

package logic;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Iterator;
import java.util.OptionalDouble;
import java.util.Random;
import settings.Settings;
import view.Chart;
import view.MainWindow;
import view.ProportionRow;
import view.ResultRaport;
import view.SimulationWindow;

public class Simulation extends Thread {

	/**
	 * Liczniki osobnikw.
	 */
	private int populationAmount;
	private int thiefsAmount;
	private int producersAmount;
	private int mutatedThiefsAmount;
	private int mutatedProducersAmount;
	
	/**
	 * Liczniki okrelajce ile mutantw danego typu wprowadzono rcznie.
	 */
	private int thiefsMutationTries;
	private int producerMutationTries;

	/**
	 * Flagi zwizane z kontrol symulacji.
	 */
	private boolean active;
	public volatile boolean flag;

	/**
	 * Odwoanie do okna symulacji.
	 */
	private SimulationWindow window;

	/**
	 * Licznik plikw - zwizany z logami.
	 */
	private int fileCoutner;

	/**
	 * Dugo aktualnego cyklu
	 */
	public int lengthOfCycle;

	/**
	 * Limity warto w macierzy wypat.
	 */
	private long thiefsMaxValue;
	private long producersMaxValue;
	private long thiefsMinValue;
	private long producersMinValue;

	/**
	 * Numer aktualnie trwajcego cyklu.
	 */
	private long numberOfCycle = 0;

	/**
	 * Czasy pocztku i koca symulacji.
	 */
	private Calendar startTime;
	private Calendar stopTime;

	/**
	 * Odwoania do wykresw.
	 */
	private Chart meanChart;
	private Chart chartThief;
	private Chart chartProducers;

	/**
	 * rednie wartoci liczebnoci osobnikw danej strategii w trakcie symulacji.
	 */
	private OptionalDouble averageThiefs;
	private OptionalDouble averageProducers;
	private OptionalDouble averageMutatedThiefs;
	private OptionalDouble averageMutatedProducers;


	/**
	 * Listy zawierajce dane o historii symulacji.
	 */
	private ArrayList<Integer> thiefsAmountList ;
	private ArrayList<Integer> producersAmountList ;
	private ArrayList<Integer> thiefsMutatedAmountList ;
	private ArrayList<Integer> producersMutatedAmountList ;

	/**
	 * Generator liczb pseudolosowych.
	 */
	private Random generator = new Random();

	/**
	 * Zmienne zwizane z tworzeniem logw.
	 */
	private StringBuilder loger = new StringBuilder();
	private int logCount = 0;

	/**
	 * Numer aktualnej rundy.
	 */
	private int round = 0;

	/**
	 * Konstruktor klasy symulacja. Inicjalizuje podstawowe pola.
	 * @param simulationWindow Odniesienie do okna symulacji
	 */
	public Simulation(SimulationWindow simulationWindow) {
		window = simulationWindow;

		active = true;
		flag = true;

		thiefsAmountList = new ArrayList<Integer>();
		producersAmountList = new ArrayList<Integer>();
		thiefsMutatedAmountList = new ArrayList<Integer>();
		producersMutatedAmountList = new ArrayList<Integer>();

		populationAmount = window.getThiefsAmount() + window.getProducersAmount();
		thiefsAmount = window.getThiefsAmount();
		producersAmount = window.getProducersAmount();

		thiefsMaxValue = thiefsAmount;
		producersMaxValue = producersAmount;
		thiefsMinValue = thiefsAmount;
		producersMinValue = producersAmount;

		thiefsAmountList.add(new Integer(thiefsAmount));
		producersAmountList.add(new Integer(producersAmount));
		thiefsMutatedAmountList.add(new Integer(0));
		producersMutatedAmountList.add(new Integer(0));

		startTime = Calendar.getInstance();
	}

	/**
	 * Metoda losujca dugo nastpnego cyklu.
	 */
	public void calculateNewSelectionDayFrequency()
	{
		lengthOfCycle = generator.nextInt(window.getFrequencyTo() - window.getFrequencyFrom() + 1) + window.getFrequencyFrom();
	}


	/**
	 * Gowna metoda programu. To tutaj znajduje si ptla czasu.
	 */
	public void run() {
		if(Settings.isLog())
			createDir();

		/**
		 * Tworzenie populacji.
		 */
		ArrayList<PopulationMember> population = new ArrayList<PopulationMember>(populationAmount);
		for (int i = 0; i < thiefsAmount; i++) {
			population.add(new Thief(window.getThiefVsThief(), window.getThiefVsProducer()));
		}
		for (int i = 0; i < producersAmount; i++) {
			population.add(new Producer(window.getProducerVsProducer(), window.getProducerVsThief()));
		}

		calculateNewSelectionDayFrequency();

		while (flag) {
			if (active) {
				if (numberOfCycle == 0)
					calculateResult(population);

				singleGameMeeting(population);

				int howMuchDying = (int) (population.size() * 0.01
						* (100 - Integer.valueOf(window.getStrengthSelection())));

				Collections.sort(population);
				makeRaport(population, howMuchDying);

				calculatePopulationStructure(population, round + 1);

				try {
					Thread.sleep(Long.valueOf(window.getDelayTextField().getText()));
				} catch (InterruptedException e1) {
					e1.printStackTrace();
				}

				round++;

				if (round == lengthOfCycle) {
					selectionAndDuplication(population);
					calculateResult(population);
					calculateNewSelectionDayFrequency();
					round = 0;
				}
			} else
				try {
					Thread.sleep(100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
		}
	}

	/**
	 * Obliczanie wartoci pokazywanych na wykresach
	 */
	private void calculateResult(ArrayList<PopulationMember> population) {

		/**
		 * Liczenie osobnikw.
		 */
		thiefsAmount = (int) population.parallelStream().filter(member -> member instanceof Thief).count();
		producersAmount= (int) population.parallelStream().filter(member -> member instanceof Producer).count();
		mutatedThiefsAmount= (int) population.parallelStream().filter(member -> member instanceof MutatedThief).count();
		mutatedProducersAmount= (int) population.parallelStream().filter(member -> member instanceof MutatedProducer).count();
		/**
		 * Dodanie informacji o liczbie osobnikw w danym cyklu do historii.
		 */
		thiefsAmountList.add(new Integer(thiefsAmount));
		producersAmountList.add(new Integer(producersAmount));
		thiefsMutatedAmountList.add(new Integer(mutatedThiefsAmount));
		producersMutatedAmountList.add(new Integer(mutatedProducersAmount));
		/**
		 * Obliczenie wartoci rednich.
		 */
		averageThiefs = thiefsAmountList.stream().mapToDouble(a -> a).average();
		averageProducers = producersAmountList.stream().mapToDouble(a -> a).average();
		averageMutatedThiefs = thiefsMutatedAmountList.stream().mapToDouble(a -> a).average();
		averageMutatedProducers = producersMutatedAmountList.stream().mapToDouble(a -> a).average();

		/**
		 * Aktualizacja wykresw.
		 */
		if (meanChart != null)
			meanChart.addThiefValue(numberOfCycle, averageThiefs.getAsDouble());
		if (chartThief != null)
			chartThief.addNormalValue(numberOfCycle, thiefsAmount);
		if (meanChart != null)
			meanChart.addProducerValue(numberOfCycle, averageProducers.getAsDouble());
		if (chartProducers != null)
			chartProducers.addNormalValue(numberOfCycle, producersAmount);
		if (meanChart != null)
			meanChart.addMutatedThiefValue(numberOfCycle, averageMutatedThiefs.getAsDouble());
		if (chartThief != null)
		{
			if(mutatedThiefsAmount > 0 && mutatedThiefsAmount < getMinimumAmmountOfMutatetToShow())
				chartThief.addMutatedValue(numberOfCycle, getMinimumAmmountOfMutatetToShow());
			else
				chartThief.addMutatedValue(numberOfCycle, mutatedThiefsAmount);
		}
		if (meanChart != null)
			meanChart.addMutatedProducerValue(numberOfCycle, averageMutatedProducers.getAsDouble());
		if (chartProducers != null)
		{
			if(mutatedProducersAmount > 0 && mutatedProducersAmount < getMinimumAmmountOfMutatetToShow())
				chartProducers.addMutatedValue(numberOfCycle, getMinimumAmmountOfMutatetToShow());
			else
				chartProducers.addMutatedValue(numberOfCycle, mutatedProducersAmount);
		}

		writeLog("STATUS: (P: " + population.size()+ " T: " + thiefsAmount + " P: " + producersAmount + " MT: " + mutatedThiefsAmount + " MP: " + mutatedProducersAmount + ")");


		numberOfCycle++;

		/**
		 * Aktualizacja statystyk.
		 */
		if (mutatedThiefsAmount == 0 && mutatedProducersAmount == 0) {
			if (thiefsAmount > thiefsMaxValue)
				thiefsMaxValue = thiefsAmount;
			if (thiefsAmount < thiefsMinValue)
				thiefsMinValue = thiefsAmount;
			if (producersAmount > producersMaxValue)
				producersMaxValue = producersAmount;
			if (producersAmount < producersMinValue)
				producersMinValue = producersAmount;
		}
	}

	/**
	 * Obliczanie minimalnego wskazania wykresu w przypadku obecnoci zmutowanych osobnikw. (Istotne gdy chcemy zabserwowa prby pojawienia si mutantw w populacji liczcej duo osobnikw).
	 */
	private int getMinimumAmmountOfMutatetToShow()
	{
		if(populationAmount>100)
			return populationAmount/100;
		else
			return 1;
	}

	
	private void singleGameMeeting(ArrayList<PopulationMember> population) {

		paintProgressBars();

		/**
		 * Wartoci tymaczasowe zawierajce kwoty z macierzy wypat.
		 */
		int tvt = 0, pvp = 0, pvt = 0;

		Collections.shuffle(population);
		for (int i = 0; i < populationAmount - 1; i += 2) {
			population.get(i).fight(population.get(i + 1));
			population.get(i + 1).fight(population.get(i));
			if(population.get(i) instanceof Thief && population.get(i+1) instanceof Thief)
				tvt++;
			else if(population.get(i) instanceof Producer && population.get(i+1) instanceof Producer)
				pvp++;
			else
				pvt++;
		}

		writeLog("SGM: (TvT: " + tvt + " PvP: " + pvp + " PvT: " + pvt + ")");
	}

	/**
	 * Metoda odwieajca paski z liczebnocia osobnikw poszczeglnych populacji.
	 */
	private void paintProgressBars()
	{
		window.getThiefsProgressBar().setValue(thiefsAmount);
		window.getProducersProgressBar().setValue(producersAmount);
		window.getTextFieldThief().setText(String.valueOf(thiefsAmount));
		window.getTextFieldProducer().setText(String.valueOf(producersAmount));
		window.getMutatedThiefsProgressBar().setValue(mutatedThiefsAmount);
		window.getMutatedProducersProgressBar().setValue(mutatedProducersAmount);
		window.getTextFieldMutatedThief().setText(String.valueOf(mutatedThiefsAmount));
		window.getTextFieldMutatedProducer().setText(String.valueOf(mutatedProducersAmount));
		window.repaint();
	}

	/**
	 * Metoda obliczajca procentowy udzia poszczeglnych strategii w populacji w danej rundzie.
	 */
	private void calculatePopulationStructure(ArrayList<PopulationMember> population, int round)
	{
		int howMuchDying = (int) (population.size() * 0.01 * (100 - Integer.valueOf(window.getStrengthSelection())));
		Collections.sort(population);
		ArrayList<PopulationMember> clone = new ArrayList<PopulationMember>(
				population.subList(howMuchDying, population.size()));

		int thiefAmmount = 0;
		int producerAmmount = 0;
		int mutatedThiefAmmount = 0;
		int mutatedProducersAmmount = 0;

		for (PopulationMember person : clone) {
			if (person instanceof MutatedThief)
				mutatedThiefAmmount++;
			else if (person instanceof Thief)
				thiefAmmount++;
			else if (person instanceof MutatedProducer)
				mutatedProducersAmmount++;
			else if (person instanceof Producer)
				producerAmmount++;
		}
		window.preparePopulationProportionPane(new ProportionRow(
				((double) thiefAmmount) / ((double) clone.size()),
				((double) producerAmmount) / ((double) clone.size()),
				((double) mutatedThiefAmmount) / ((double) clone.size()),
				((double) mutatedProducersAmmount) / ((double) clone.size()),
				(round == lengthOfCycle),
				numberOfCycle),
				population,
				population.size() - howMuchDying
				);
	}

	/**
	 * Realizacja selekcji i duplikacji.
	 * @param population Populacja
	 */
	private void selectionAndDuplication(ArrayList<PopulationMember> population) {
		int howMuchDying = (int) (population.size() * 0.01
				* (100 - Integer.valueOf(window.getStrengthSelection())));

		Collections.sort(population);

		for (PopulationMember mem : population) {
			mem.setPoints(Params.STARTING_SPICES_POINTS);
		}

		ArrayList<PopulationMember> clone = new ArrayList<PopulationMember>(
				population.subList(howMuchDying, population.size()));
		Collections.shuffle(clone);
		Iterator<PopulationMember> itr;
		int tempCnt = 0;
		PopulationMember temp = null;
		while (tempCnt < howMuchDying) {
			itr = clone.iterator();
			while (itr.hasNext()) {
				if (tempCnt == howMuchDying)
					break;
				temp = itr.next();
				if (temp instanceof Thief)
					population.set(tempCnt, new Thief(window.getThiefVsThief(), window.getThiefVsProducer()));
				else if (temp instanceof Producer)
					population.set(tempCnt, new Producer(window.getProducerVsProducer(), window.getProducerVsThief()));
				else if (temp instanceof MutatedThief)
					population.set(tempCnt,
							new MutatedThief(window.getMutateThiefVsThief(), window.getMutateThiefVsProducer()));
				else if (temp instanceof MutatedProducer)
					population.set(tempCnt,
							new MutatedProducer(window.getMutateProducerVsProducer(), window.getMutateProducerVsThief()));
				tempCnt++;
			}
		}
		if (window.isThiefMutate() && thiefsAmount!=0) {
			for(int i = 0; i < population.size(); i++)
				if (population.get(i) instanceof Thief)
				{
					population.set(i, new MutatedThief(window.getMutateThiefVsThief(), window.getMutateThiefVsProducer()));
					break;
				}
			window.setThiefMutate(false);
			window.getThiefMutationAmountInfo().setText(window.getThiefMutationAmountInfo().getText().replaceAll("\\d+", "") +String.valueOf(++thiefsMutationTries));
			mutatedThiefsAmount = 1;
		}
		if (window.isProducerMutate() && producersAmount!=0) {
			for(int i = 0; i < population.size(); i++)
				if (population.get(i) instanceof Producer)
				{
					population.set(i, new MutatedProducer(window.getMutateProducerVsProducer(), window.getMutateProducerVsThief()));
					break;
				}
			window.setProducerMutate(false);
			window.getProducerMutationAmountInfo().setText(window.getProducerMutationAmountInfo().getText().replaceAll("\\d+", "") +String.valueOf(++producerMutationTries));
			mutatedProducersAmount = 1;
		}
	}
	
	/**
	 * Stworzeni plikw zawierajcych informacje o strukturze populacji w pierwszych 50 cyklach.
	 */
	public void makeRaport(ArrayList<PopulationMember> population, int howMuchDying) {
		if(!Settings.isLog())
			return;

		StringBuilder builder = new StringBuilder();
		for (int i = 0; i < populationAmount; i++) {
			builder.append(
					String.valueOf(i) + "  =  " + population.get(i).toString() + System.getProperty("line.separator"));
			if (i == howMuchDying - 1)
				builder.append(Params.SELECTION_SEPARATOR
						+ System.getProperty("line.separator"));
		}
		PrintWriter writer = null;
		String path = Params.DIR_NAME + MainWindow.dirCounter + "/" + Params.FILE_NAME + fileCoutner + ".txt";
		if (fileCoutner < 3500)
			try {
				writer = new PrintWriter(new FileOutputStream(new File(path),
						true));
				writer.append(builder.toString());

			} catch (IOException e) {
				e.printStackTrace();
			} finally {
				writer.close();
			}
		fileCoutner++;
	}

	/**
	 * Stworzenie folderu z raportami.
	 */
	private void createDir()
	{
		/**
		 * Dla kadej nowej symulacji twrz folder "reportNoX" gdzie bd znajdowa si logi na jej temat.
		 */
		File dir = new File(Params.DIR_NAME + MainWindow.dirCounter);
		if (dir.exists())
			deleteDir(dir);
		dir.mkdir();
	}
	
	/**
	 * Usu folder z raportami.
	 * @param dir Scieka do folderu z raportami.
	 */
	public static void deleteDir(File dir) {
		File[] files = dir.listFiles();

		if(files == null)
			return;

		for (File myFile : files) {
			if (myFile.isDirectory()) {
				deleteDir(myFile);
			}
			myFile.delete();
		}
	}

	/**
	 * Inicjalizuj wykresy
	 * @param chart
	 * @param chartThief
	 * @param chartProducer
	 */
	public void startCharting(Chart chart, Chart chartThief, Chart chartProducer) {
		this.meanChart = chart;
		this.chartThief = chartThief;
		this.chartProducers = chartProducer;
	}

	/**
	 * Wypenij okienko z raportem.
	 * @param raport Odniesienie do okna z raportem.
	 */
	public void raportToFrame(ResultRaport raport) {
		DateFormat dateFormat = new SimpleDateFormat("dd/MM/yyyy");
		stopTime = Calendar.getInstance();
		raport.dayField.setText(dateFormat.format(stopTime.getTime()));

		SimpleDateFormat timeFormat = new SimpleDateFormat("HH:mm:ss.SS");
		raport.timeStartField.setText(timeFormat.format(startTime.getTime()));
		stopTime = Calendar.getInstance();
		raport.timeEndField.setText(timeFormat.format(stopTime.getTime()));

		raport.thiefAmountField.setText(String.valueOf(window.getThiefsAmount()));
		raport.producerAmountField.setText(String.valueOf(window.getProducersAmount()));
		raport.populationAmountField.setText(String.valueOf(window.getThiefsAmount() + window.getProducersAmount()));
		raport.thiefBothRatioField.setText(String.valueOf((int)thiefsAmountList.stream().mapToDouble(a -> a).average().getAsDouble()));
		raport.producersBothRatioField.setText(String.valueOf((int)producersAmountList.stream().mapToDouble(a -> a).average().getAsDouble()));
		raport.percentBothThiefsField
				.setText(String.format("%.2f", (thiefsAmountList.stream().mapToDouble(a -> a).average().getAsDouble()) / populationAmount * 100));
		raport.percentBothProducersField
				.setText(String.format("%.2f", (100d -(thiefsAmountList.stream().mapToDouble(a -> a).average().getAsDouble()) / populationAmount * 100)));
		raport.thiefMaxRatioField.setText(String.valueOf(thiefsMaxValue));
		raport.producersMaxRatioField.setText(String.valueOf(producersMaxValue));
		raport.thiefMinRatioField.setText(String.valueOf(thiefsMinValue));
		raport.producersMinRatioField.setText(String.valueOf(producersMinValue));
		raport.percentMaxThiefField.setText(String.format("%.2f", (double) thiefsMaxValue / populationAmount * 100));
		raport.percentMinThiefField.setText(String.format("%.2f", (double) thiefsMinValue / populationAmount * 100));
		raport.percentMaxProducersField.setText(String.format("%.2f", (double) producersMaxValue / populationAmount * 100));
		raport.percentMinProducersField.setText(String.format("%.2f", (double) producersMinValue / populationAmount * 100));
		raport.thiefThiefValue1.setText(String.valueOf(window.getThiefVsThief()));
		raport.thiefThiefValue2.setText(String.valueOf(window.getThiefVsThief()));
		raport.producerProducerValue1.setText(String.valueOf(window.getProducerVsProducer()));
		raport.producerProducerValue2.setText(String.valueOf(window.getProducerVsProducer()));
		raport.producerThiefValue1.setText(String.valueOf(window.getThiefVsProducer()));
		raport.producerThiefValue2.setText(String.valueOf(window.getProducerVsThief()));
		raport.thiefProducerValue1.setText(String.valueOf(window.getProducerVsThief()));
		raport.thiefProducerValue2.setText(String.valueOf(window.getThiefVsProducer()));
	}

	public void setActive(boolean active) {
		this.active = active;
	}

	public boolean isActive() {
		return active;
	}

	/**
	 * Napisz tekst do pliku z logami
	 * @param log Tre logu.
	 */
	private void writeLog(String log) {
		if(!Settings.isLog())
			return;

		ZonedDateTime time = ZonedDateTime.now();
		DateTimeFormatter formatter = DateTimeFormatter.ofPattern("hh:mm");

		loger.append(time.format(formatter) + " " + numberOfCycle + ":" + round + " ");
		loger.append(log);
		loger.append(System.getProperty("line.separator"));
		logCount++;

		if(logCount >= 500 || round == lengthOfCycle-1)
		{
			logCount = 0;
			saveLog();
		}

	}
	
	/**
	 * Przepisz wytworzone logi z pamici do pliku.
	 */
	private void saveLog()
	{
		PrintWriter writer = null;
		String path = Params.DIR_NAME + MainWindow.dirCounter + "/" + Params.LOG_FILE_NAME;
		try {
			writer = new PrintWriter(new FileOutputStream(new File(path),
					false));
			writer.append(loger.toString());

		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			writer.close();
		}
	}
}
